# -*- coding: utf-8 -*-
"""
Created on Thu Apr 22 21:09:41 2021

@author: Peter
"""
from typing import Optional, List
from dataclasses import dataclass
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import pandas as pd
import numpy as np
from numpy.random import normal
from zombie.weapon_stats import _weapon_stats_dic, _gun_names

rare_level = {'common': 1.00,
              'green': 1.50,
              'blue': 3.00,
              'purple': 6.00,
              'orange': 12.00,
              }

pack_level = {'0': 1.00,
              '1': 2.00,
              '2': 4.00,
              '3': 8.00,
              }

_weapon_type_dict = {'Launcher': {'elites': 1.00,
                                  'Armour Damage': 1.00,
                                  'ammo': 1.00},
                     'Special': {'elites': 1.00,
                                 'Armour Damage': 1.00,
                                 'ammo': 1.00},
                     'Smg': {'Close': 1.00,
                             'Crit': 1.00,
                             'Attachments': 5},
                     'Shotgun': {'Close': 1.00,
                                 'Crit': 1.00,
                                 'Armour Damage': 1.00},
                     'Pistol': {'Close': 1.00,
                                'Crit': 1.00,
                                'Armour Damage': 1.00},
                     'Marksman': {'Long': 1.00,
                                  'Crit': 1.00,
                                  'Attachments': 5},
                     'Sniper': {'Armour Damage': 1.00,
                                'Crit': 1.00,
                                'Attachments': 5},
                     'Lmg': {'Armour Damage': 1.00,
                             'Crit': 1.00,
                             'Attachments': 5},
                     'Assault': {'Long': 1.00,
                                 'Crit': 1.00,
                                 'Attachments': 5},
                     'Melee': {'object': 'gun butt',
                               'Damage': 1.00,
                               'regen': 0}
                     }

_perk_type_dict = {'speed': {'swap': 0,
                             'field recharge': 0,
                             'reload': 1.15,
                             'barr. and myst.': 0,
                             'fire and reload': 0},
                   'stamin up': {'movement': 1.07,
                                 'sprint': 1.07,
                                 'back pedal': 1.00,
                                 'fall damage': 0,
                                 'aim walking': 1.00,
                                 'equipment': 1.00,
                                 'sprint falloff': 0},
                   'deadshot': {'scope sway': 0,
                                'critical1': 1.00,
                                'armour': 1.00,
                                'hip fire spread': 1.00,
                                'critical': 1.00,
                                'consecutive': 1},
                   'death_perception': {'minimap': 0,
                                        'indicators': 0,
                                        'Salvage': 1.00,
                                        'Armour Damage': 1.00,
                                        'chests': 0},
                   }


def _shot_distribution(stats) -> np.ndarray:
    acc = [val.gun_acc_value for val in stats]
    acc_n = [val for val in acc if val != 0]
    sig1 = np.around(np.std(acc_n), 3)
    mu1 = np.around(np.mean(acc_n), 3)
    return np.around(normal(mu1, sig1, 2500), 3)


def _set_weapon_class(weapon_type: str, weapon_tier: str) -> dict:
    if weapon_type == 'Launcher':
        return {'1': {'elites': 1.10, 'Armour Damage': 1.00, 'ammo': 1.00},
                '2': {'elites': 1.10, 'Armour Damage': 1.10, 'ammo': 1.00},
                '3': {'elites': 1.10, 'Armour Damage': 1.10, 'ammo': 0.75},
                '4': {'elites': 1.25, 'Armour Damage': 1.10, 'ammo': 0.75},
                '5': {'elites': 1.25, 'Armour Damage': 1.25, 'ammo': 0.75}
                }[weapon_tier]
    elif weapon_type == 'Special':
        return {'1': {'elites': 1.10, 'Armour Damage': 1.00, 'ammo': 1.00},
                '2': {'elites': 1.10, 'Armour Damage': 1.10, 'ammo': 1.00},
                '3': {'elites': 1.10, 'Armour Damage': 1.10, 'ammo': 0.75},
                '4': {'elites': 1.25, 'Armour Damage': 1.10, 'ammo': 0.75},
                '5': {'elites': 1.25, 'Armour Damage': 1.25, 'ammo': 0.75}
                }[weapon_tier]
    elif weapon_type == 'Smg':
        return {'1': {'Close': 1.10, 'Crit': 1.00, 'Attachments': 1},
                '2': {'Close': 1.10, 'Crit': 1.10, 'Attachments': 1},
                '3': {'Close': 1.10, 'Crit': 1.10, 'Attachments': 1.6},
                '4': {'Close': 1.25, 'Crit': 1.10, 'Attachments': 1.6},
                '5': {'Close': 1.25, 'Crit': 1.25, 'Attachments': 1.6}
                }[weapon_tier]
    elif weapon_type == 'Shotgun':
        return {'1': {'Close': 1.10, 'Crit': 1.00, 'Armour Damage': 1.00},
                '2': {'Close': 1.10, 'Crit': 1.10, 'Armour Damage': 1.00},
                '3': {'Close': 1.10, 'Crit': 1.10, 'Armour Damage': 1.10},
                '4': {'Close': 1.25, 'Crit': 1.10, 'Armour Damage': 1.10},
                '5': {'Close': 1.25, 'Crit': 1.25, 'Armour Damage': 1.10}
                }[weapon_tier]
    elif weapon_type == 'Pistol':
        return {'1': {'Close': 1.10, 'Crit': 1.00, 'Armour Damage': 1.00},
                '2': {'Close': 1.10, 'Crit': 1.10, 'Armour Damage': 1.00},
                '3': {'Close': 1.10, 'Crit': 1.10, 'Armour Damage': 1.10},
                '4': {'Close': 1.25, 'Crit': 1.10, 'Armour Damage': 1.10},
                '5': {'Close': 1.25, 'Crit': 1.25, 'Armour Damage': 1.10}
                }[weapon_tier]
    elif weapon_type == 'Marksman':
        return {'1': {'Long': 1.10, 'Crit': 1.00, 'Attachments': 1},
                '2': {'Long': 1.10, 'Crit': 1.10, 'Attachments': 1},
                '3': {'Long': 1.10, 'Crit': 1.10, 'Attachments': 1.6},
                '4': {'Long': 1.25, 'Crit': 1.10, 'Attachments': 1.6},
                '5': {'Long': 1.25, 'Crit': 1.25, 'Attachments': 1.6}
                }[weapon_tier]
    elif weapon_type == 'Sniper':
        return {'1': {'Armour Damage': 1.10, 'Crit': 1.00, 'Attachments': 1},
                '2': {'Armour Damage': 1.10, 'Crit': 1.10, 'Attachments': 1},
                '3': {'Armour Damage': 1.10, 'Crit': 1.10, 'Attachments': 1.6},
                '4': {'Armour Damage': 1.25, 'Crit': 1.10, 'Attachments': 1.6},
                '5': {'Armour Damage': 1.25, 'Crit': 1.25, 'Attachments': 1.6}
                }[weapon_tier]
    elif weapon_type == 'Lmg':
        return {'1': {'Armour Damage': 1.10, 'Crit': 1.00, 'Attachments': 1},
                '2': {'Armour Damage': 1.10, 'Crit': 1.10, 'Attachments': 1},
                '3': {'Armour Damage': 1.10, 'Crit': 1.10, 'Attachments': 1.6},
                '4': {'Armour Damage': 1.25, 'Crit': 1.10, 'Attachments': 1.6},
                '5': {'Armour Damage': 1.25, 'Crit': 1.25, 'Attachments': 1.6}
                }[weapon_tier]
    elif weapon_type == 'Assault':
        return {'1': {'Long': 1.10, 'Crit': 1.00, 'Attachments': 1},
                '2': {'Long': 1.10, 'Crit': 1.10, 'Attachments': 1},
                '3': {'Long': 1.10, 'Crit': 1.10, 'Attachments': 1.6},
                '4': {'Long': 1.25, 'Crit': 1.10, 'Attachments': 1.6},
                '5': {'Long': 1.25, 'Crit': 1.25, 'Attachments': 1.6}
                }[weapon_tier]
    elif weapon_type == 'Melee':
        return {'1': {'object': 'knife', 'Damage': 1.00, 'regen': 0},
                '2': {'object': 'knife', 'Damage': 1.10, 'regen': 0},
                '3': {'object': 'bowie', 'Damage': 1.10, 'regen': 0},
                '4': {'object': 'bowie', 'Damage': 1.25, 'regen': 0},
                '5': {'object': 'bowie', 'Damage': 1.25, 'regen': 1}
                }[weapon_tier]


def _set_perk_class(perk_type: str, perk_tier: str) -> dict:
    if perk_type == 'speed':
        return {'1': {'swap': 1, 'field recharge': 1, 'Reload': 0.85, 'barr. and myst.': 0,
                      'Shooting Speed': 1.00},
                '2': {'swap': 1, 'field recharge': 0.80, 'Reload': 0.85, 'barr. and myst.': 0,
                      'Shooting Speed': 1.00},
                '3': {'swap': 1, 'field recharge': 0.80, 'Reload': 0.70, 'barr. and myst.': 0,
                      'Shooting Speed': 1.00},
                '4': {'swap': 1, 'field recharge': 0.80, 'Reload': 0.70, 'barr. and myst.': 1,
                      'Shooting Speed': 1.00},
                '5': {'swap': 1, 'field recharge': 0.80, 'Reload': 0.70, 'barr. and myst.': 1,
                      'Shooting Speed': 1.07}
                }[perk_tier]
    elif perk_type == 'stamin up':
        return {'1': {'Movement Speed': 1.07, 'Sprinting Speed': 1.07, 'back pedal': 1.07,
                      'fall damage': 0, 'Aim Walking': 1.00, 'equipment': 1.00, 'sprint falloff': 0,
                      'Shooting Speed': 1.07},
                '2': {'Movement Speed': 1.07, 'Sprinting Speed': 1.07, 'back pedal': 1.07,
                      'fall damage': 1, 'Aim Walking': 1.00, 'equipment': 1.00, 'sprint falloff': 0,
                      'Shooting Speed': 1.07},
                '3': {'Movement Speed': 1.07, 'Sprinting Speed': 1.07, 'back pedal': 1.07,
                      'fall damage': 1, 'Aim Walking': 1.07, 'equipment': 1.00, 'sprint falloff': 0,
                      'Shooting Speed': 1.07},
                '4': {'Movement Speed': 1.07, 'Sprinting Speed': 1.07, 'back pedal': 1.07,
                      'fall damage': 1, 'Aim Walking': 1.07, 'equipment': 1.07, 'sprint falloff': 0,
                      'Shooting Speed': 1.07},
                '5': {'Movement Speed': 1.07, 'Sprinting Speed': 1.07, 'back pedal': 1.07,
                      'fall damage': 1, 'Aim Walking': 1.07, 'equipment': 1.07, 'sprint falloff': 1,
                      'Shooting Speed': 1.07}
                }[perk_tier]
    elif perk_type == 'deadshot':
        return {'1': {'Idle Sway': 0, 'critical1': 2.00, 'Armour Damage': 1.00, 'Hip Fire': 1.00,
                      'Crit': 1.00, 'consecutive': 0},
                '2': {'Idle Sway': 0, 'critical1': 2.00, 'Armour Damage': 1.50, 'Hip Fire': 1.00,
                      'Crit': 1.00, 'consecutive': 0},
                '3': {'Idle Sway': 0, 'critical1': 2.00, 'Armour Damage': 1.50, 'Hip Fire': 0.70,
                      'Crit': 1.00, 'consecutive': 0},
                '4': {'Idle Sway': 0, 'critical1': 2.00, 'Armour Damage': 1.50, 'Hip Fire': 0.70,
                      'Crit': 1.10, 'consecutive': 0},
                '5': {'Idle Sway': 0, 'critical1': 2.00, 'Armour Damage': 1.50, 'Hip Fire': 0.70,
                      'Crit': 1.10, 'consecutive': 1}
                }[perk_tier]
    elif perk_type == 'death_perception':
        return {'1': {'minimap': 0, 'indicators': 0, 'Salvage': 1.00, 'Armour Damage': 1.00, 'chests': 0},
                '2': {'minimap': 0, 'indicators': 0, 'Salvage': 1.00, 'Armour Damage': 1.00, 'chests': 0},
                '3': {'minimap': 0, 'indicators': 0, 'Salvage': 1.20, 'Armour Damage': 1.00, 'chests': 0},
                '4': {'minimap': 0, 'indicators': 0, 'Salvage': 1.20, 'Armour Damage': 1.25, 'chests': 0},
                '5': {'minimap': 0, 'indicators': 0, 'Salvage': 1.20, 'Armour Damage': 1.25, 'chests': 0}
                }[perk_tier]


def _apply_multipliers(weapon_multi: dict, perk_multi: dict, weapon_dic: dict) -> dict:
    dn = weapon_dic.copy()
    # Weapons
    for j in weapon_multi[dn['Weapon Type']]:
        dn[j] = round(dn[j] * weapon_multi[dn['Weapon Type']][j], 2)

    # Perks
    for i in dn.keys():
        for j in perk_multi.keys():
            for k in perk_multi[j].keys():
                if i == k:
                    dn[i] = round(dn[i] * perk_multi[j][k], 2)

    return dn


def _apply_attachments(weapon_dic: dict, attachment_lst: List[str]) -> dict:

    if weapon_dic['Pack'] != '0':
        pack_mag = True
    else:
        pack_mag = False

    dn = weapon_dic.copy()
    dm = {i: 0.0 for i in weapon_dic.keys()}
    for i in attachment_lst:
        if i == 'None':
            continue
        else:
            at = i.split('%')[0]
            at_value = at.split(' ')[1].strip()
            at_pos = at.split(' ')[0].strip()
            val = float(at_value) / 100

        if '% Damage' in i:
            dm['Damage'] += val
            dm['Damage 2'] += val
            dm['Damage 3'] += val
            continue
        elif 'Effective Damage Range' in i:
            if at_pos == '+':
                dm['Range'] += val
                dm['Range 2'] += val
                continue
            else:
                dm['Range'] -= val
                dm['Range 2'] -= val
                continue
        elif 'Rate of Fire' in i:
            if at_pos == '+':
                dm['Fire Rate'] += val
                continue
            else:
                dm['Fire Rate'] -= val
                continue
        elif 'Fire Rate' in i:
            if at_pos == '+':
                dm['Fire Rate'] += val
                continue
            else:
                dm['Fire Rate'] -= val
                continue
        elif 'Bullet Velocity' in i:
            if at_pos == '+':
                dm['Velocity'] += val
                continue
            else:
                dm['Velocity'] -= val
                continue
        elif 'Armour Damage' in i:
            dm['Armour Damage'] += val
            continue
        elif 'Melee Quickness' in i:
            dm['Melee Quickness'] -= val
            continue
        elif 'Sprinting Move Speed' in i:
            if at_pos == '+':
                dm['Sprinting Speed'] += val
                continue
            else:
                dm['Sprinting Speed'] -= val
                continue
        elif 'Shooting Move Speed' in i:
            if at_pos == '+':
                dm['Shooting Speed'] += val
                continue
            else:
                dm['Shooting Speed'] -= val
                continue
        elif 'Sprint to Fire Time' in i:
            if at_pos == '+':
                dm['Sprint to Fire'] -= val
                continue
            else:
                dm['Sprint to Fire'] += val
                continue
        elif 'Aim Walking Movement Speed' in i:
            if at_pos == '+':
                dm['Aim Walking'] += val
                continue
            else:
                dm['Aim Walking'] -= val
                continue
        elif 'Movement Speed' in i:
            if at_pos == '+':
                dm['Movement Speed'] += val
                continue
            else:
                dm['Movement Speed'] -= val
                continue
        elif 'Aim Down Sight Time' in i:
            if at_pos == '-':
                dm['ADS'] += val
                continue
            else:
                dm['ADS'] -= val
                continue
        elif 'Vertical Recoil Control' in i:
            if at_pos == '+':
                dm['Vertical Recoil'] -= val
                continue
            else:
                dm['Vertical Recoil'] += val
                continue
        elif 'Horizontal Recoil Control' in i:
            if at_pos == '+':
                dm['Horizontal Recoil'] -= val
                continue
            else:
                dm['Horizontal Recoil'] += val
                continue
        elif 'Flinch Resistance' in i:
            if at_pos == '+':
                dm['Flinch'] -= val
                continue
            else:
                dm['Flinch'] += val
                continue
        elif 'Hip Fire Accuracy' in i:
            if at_pos == '+':
                dm['Hip Fire'] -= val
                continue
            else:
                dm['Hip Fire'] += val
                continue
        elif 'Magazine Ammo Capacity' in i:
            if at_pos == '+':
                dm['Mag Capacity'] += val
                continue
            else:
                dm['Mag Capacity'] -= val
                continue
        elif 'Reload Quickness' in i:
            if at_pos == '+':
                dm['Reload'] -= val
                continue
            else:
                dm['Reload'] += val
                continue
        elif 'Max Starting Ammo' in i:
            if at_pos == '+':
                dm['Starting Ammo'] += val
                continue
            else:
                dm['Starting Ammo'] -= val
                continue
        elif '% Ammo Capacity' in i:
            if at_pos == '+':
                dm['Ammo Capacity'] += val
                continue
            else:
                dm['Ammo Capacity'] -= val
                continue
        elif 'Increased Salvage Drop Rate' in i:
            dm['Salvage'] += val
            continue
        elif 'Increased Equipment Drop Chance' in i:
            dm['Equipment'] += val
            continue
        else:
            print('Attachment Processor Missed: ', i, at_value, at_pos)

    ind_lst = [
        'Damage',
        'Damage 2',
        'Damage 3',
        'Range',
        'Range 2',
        'Fire Rate',
        'Velocity',
        'Armour Damage',
        'Melee Quickness',
        'Sprinting Speed',
        'Shooting Speed',
        'Sprint to Fire',
        'Aim Walking',
        'Movement Speed',
        'ADS',
        'Vertical Recoil',
        'Horizontal Recoil',
        'Flinch',
        'Hip Fire',
        'Mag Capacity',
        'Reload',
        'Starting Ammo',
        'Ammo Capacity',
        'Salvage',
        'Equipment',
    ]

    for i in dm.keys():
        if i in ind_lst:
            dn[i] = round(float(dn[i]) * (1.00 + dm[i]), 2)

    if pack_mag:
        dn['Mag Capacity'] = dn['Mag Capacity'] * 2
        dn['Ammo Capacity'] = dn['Ammo Capacity'] * 2

    return dn


def _get_attachments(weapon_dic: dict, equipped_dic: dict = None):
    location = []
    name = []
    effect = []
    for part in ['Muzzle', 'Barrel', 'Body', 'Underbarrel', 'Magazine', 'Handle', 'Stock']:
        for j in weapon_dic[part].keys():
            location.append(part)
            name.append(j)
            effect.append(weapon_dic[part][j])

    df = pd.DataFrame()
    df['location'] = location
    df['name'] = name
    df['effect'] = effect

    if equipped_dic:
        temp_lst = []
        for part in ['Muzzle', 'Barrel', 'Body', 'Underbarrel', 'Magazine', 'Handle', 'Stock']:
            if part in equipped_dic.keys():
                temp_lst.append(weapon_dic[part][equipped_dic[part]])
        return sum(temp_lst, [])
    else:
        return df


@dataclass
class DamageProfile:
    """

    Builds a DamageProfile Class, which holds weapon and perk class levels to be applied later on.

    :param weapon_class_levels: Dict of weapon types and tier level.
    :type weapon_class_levels: dict
    :param perk_class_levels: Dict of perk type and tier level.
    :type perk_class_levels: dict
    :param max_range: Maximum range for calculations, default is 100. *Optional*
    :type max_range: int
    :example:
        >>> # User inputs
        >>> weapon_class_levels = {'Launcher': '5', 'Special': '5', 'Smg': '5', 'Shotgun': '5', 'Pistol': '5',
        >>>                        'Marksman': '5', 'Sniper': '5', 'Lmg': '5', 'Assault': '5', 'Melee': '5'}
        >>> perk_class_levels = {'speed': '5', 'stamin up': '5', 'deadshot': '5', 'death_perception': '5'}
        >>> # Core Classes
        >>> damage_profile = DamageProfile(weapon_class_levels=weapon_class_levels, perk_class_levels=perk_class_levels,
        >>>                                max_range=100)
    :note: *None*

    """
    def __init__(self, weapon_class_levels: dict, perk_class_levels: dict, max_range: Optional[int]):
        self._weapon_class_levels = weapon_class_levels
        self._perk_class_levels = perk_class_levels
        self._max_range = max_range
        self._hits = _shot_distribution(stats=_weapon_stats_dic.values())

        self._weapon_type_dict = {}
        for i, j in enumerate(_weapon_type_dict.keys()):
            self._weapon_type_dict[j] = _set_weapon_class(j, self._weapon_class_levels[j])

        self._perk_dict = {}
        for i, j in enumerate(_perk_type_dict.keys()):
            self._perk_dict[j] = _set_perk_class(j, self._perk_class_levels[j])

        if perk_class_levels['deadshot'] == '5':
            self._consecutive = True
        else:
            self._consecutive = False

    def __repr__(self):
        return "DamageProfile"

    @property
    def hits(self) -> np.ndarray:
        """Returns hit value multipliers to account for accuracy and critical hit percentages."""
        return self._hits

    @property
    def get_weapon_classes(self) -> dict:
        """Returns user input weapon tier levels"""
        return self._weapon_type_dict

    @property
    def get_perk_classes(self) -> dict:
        """Returns user input perk tier levels"""
        return self._perk_dict

    @property
    def consecutive(self) -> bool:
        """Returns a bool value representing if consecutive shots used"""
        return self._consecutive

    @property
    def max_range(self) -> int:
        """Max distance for calculations"""
        return self._max_range


@dataclass
class ModifiedWeapon:
    """

    Builds a ModifiedWeapon Class, which holds default and modified weapon stats.

    :param weapon_name: Name of the weapon.
    :type weapon_name: str
    :param weapon_class_levels: Dict of weapon types and tier level.
    :type weapon_class_levels: dict
    :param perk_class_levels: Dict of perk type and tier level.
    :type perk_class_levels: dict
    :param nickname: User input name for the weapon.
    :type nickname: str
    :param equipped_attachments: Dict of weapon attachment location and attachment name.
    :type equipped_attachments: dict
    :param rarity: Rarity color.
    :type rarity: str
    :param pap: Pack a punch level.
    :type pap: str
    :param accuracy: Accuracy percent with the weapon.
    :type accuracy: float
    :param critical: Critical hit percent with the weapon.
    :type critical: float
    :example:

    :note:

    """

    def __init__(self,
                 weapon_name: str,
                 weapon_class_levels: dict,
                 perk_class_levels: dict,
                 nickname: Optional[str] = None,
                 equipped_attachments: Optional[dict] = None,
                 rarity: Optional[str] = None,
                 pap: Optional[str] = None,
                 accuracy: Optional[float] = None,
                 critical: Optional[float] = None):
        self._pap = pap
        self._rarity = rarity
        self._perk_class_levels = perk_class_levels
        self._weapon_class_levels = weapon_class_levels
        d = _weapon_stats_dic[weapon_name]
        self._default_stats = {'Name': d.name,
                               'Temp Name': d.temp_name,
                               'Damage': d.damage_1,
                               'Damage 2': d.damage_2,
                               'Damage 3': d.damage_3,
                               'Range': d.range_1,
                               'Range 2': d.range_2,
                               'Fire Rate': d.fire_rate,
                               'Velocity': d.velocity,
                               'Armour Damage': d.armour_damage,
                               'Melee Quickness': d.melee,
                               'Movement Speed': d.movement,
                               'Sprinting Speed': d.sprint,
                               'Shooting Speed': d.shoot_speed,
                               'Sprint to Fire': d.sprint_to_fire,
                               'Aim Walking': d.aim_walking,
                               'ADS': d.ads,
                               'Vertical Recoil': d.vertical_recoil,
                               'Horizontal Recoil': d.horizontal_recoil,
                               'Centering Speed': d.centering_speed,
                               'Idle Sway': d.idle_sway,
                               'Flinch': d.flinch,
                               'Hip Fire': d.hip_fire,
                               'Mag Capacity': d.mag_capacity,
                               'Reload': d.reload,
                               'Starting Ammo': d.starting_ammo,
                               'Ammo Capacity': d.ammo_capacity,
                               'Burst': d.burst,
                               'PAP Burst': d.pap_burst,
                               'Long Shot': d.long_shot,
                               'Crit': d.critical,
                               'Long': d.long,
                               'Close': d.close,
                               'Salvage': d.salvage,
                               'Equipment': d.equipment,
                               'Attachments': d.attachments,
                               'Weapon Type': d.weapon_type,
                               'Accuracy': d.gun_acc_value,
                               'Critical': d.gun_critical_value,
                               'Pack': d.pack,
                               'Rare': d.rare,
                               'Muzzle': d.muzzle,
                               'Barrel': d.barrel,
                               'Body': d.body,
                               'Underbarrel': d.under_barrel,
                               'Magazine': d.magazine,
                               'Handle': d.handle,
                               'Stock': d.stock,
                               }

        self._modified_stats = self._default_stats
        self._equipped_attachments = equipped_attachments

        if nickname is not None:
            self._modified_stats['Temp Name'] = nickname

        if accuracy is not None:
            self._modified_stats['Accuracy'] = accuracy

        if critical is not None:
            self._modified_stats['Critical'] = critical

        if rarity is not None:
            self._modified_stats['Rare'] = rarity
            for key in ['Damage', 'Damage 2', 'Damage 3']:
                self._modified_stats[key] = self._modified_stats[key] * rare_level[rarity]

        if pap is not None:
            if pap != '0':
                self._modified_stats['Pack'] = pap
                self._modified_stats['Mag Capacity'] = self._modified_stats['Mag Capacity'] / self._modified_stats['PAP Burst']
                for key in ['Damage', 'Damage 2', 'Damage 3']:
                    self._modified_stats[key] = self._modified_stats[key] * pack_level[pap] * self._modified_stats['PAP Burst']

        temp_dic = _apply_multipliers(weapon_multi=self._weapon_class_levels, perk_multi=self._perk_class_levels,
                                      weapon_dic=self._modified_stats)

        if equipped_attachments is not None:
            effect_lst = _get_attachments(weapon_dic=temp_dic, equipped_dic=equipped_attachments)
            final_dic = _apply_attachments(weapon_dic=temp_dic, attachment_lst=effect_lst)

            for part in ['Muzzle', 'Barrel', 'Body', 'Underbarrel', 'Magazine', 'Handle', 'Stock']:
                if part in equipped_attachments.keys():
                    final_dic[part] = {equipped_attachments[part]: final_dic[part][equipped_attachments[part]]}
                else:
                    final_dic[part] = {'None': 'None'}
            temp_dic = final_dic

        self._modified_stats = temp_dic
        col = self._modified_stats['Temp Name']
        index_lst = ['Armour Damage', 'Damage', 'Range', 'Velocity', 'ADS', 'Aim Walking', 'Sprint to Fire',
                     'Shooting Speed', 'Sprinting Speed', 'Movement Speed']
        self._modified_values = pd.DataFrame.from_dict(self._modified_stats,
                                                       orient='index',
                                                       columns=[col]).loc[index_lst, col]

    def __repr__(self):
        return self._default_stats['Name']

    @property
    def pack_a_punch_level(self) -> str:
        """Returns pack a punch level"""
        return self._pap

    @property
    def rarity_level(self) -> str:
        """Returns rarity level"""
        return self._rarity

    @property
    def default_stats(self) -> dict:
        """Returns default stats for the weapon **before** perks, weapon tiers and attachments"""
        return self._default_stats

    @property
    def modified_stats(self) -> dict:
        """Returns default stats for the weapon **after** perks, weapon tiers and attachments"""
        return self._modified_stats

    @property
    def get_attachments(self) -> pd.DataFrame:
        """Returns attachments and effects as a DataFrame"""
        if self._equipped_attachments:
            vals = self._modified_stats
        else:
            vals = self._default_stats

        location, name, effect = [], [], []
        for part in ['Muzzle', 'Barrel', 'Body', 'Underbarrel', 'Magazine', 'Handle', 'Stock']:
            for j in vals[part].keys():
                location.append(part)
                name.append(j)
                effect.append(self.modified_stats[part][j])

        df = pd.DataFrame()
        df['location'] = location
        df['name'] = name
        df['effect'] = effect

        if self._equipped_attachments:
            df = df.set_index('location').loc[self._equipped_attachments.keys()]
        else:
            df = df.set_index('location').sort_index()
        return df

    def ax(self):
        """Returns a Horizontal Bar Plot comparing the weapons default verse modified stats"""
        col = self._modified_stats['Name']
        index_lst = ['Armour Damage', 'Damage', 'Range', 'Velocity', 'ADS', 'Aim Walking', 'Sprint to Fire',
                     'Shooting Speed', 'Sprinting Speed', 'Movement Speed']

        # Set up
        temp_mod = self._modified_values
        temp_def = pd.DataFrame.from_dict(self._default_stats, orient='index', columns=[col]).loc[index_lst, col]
        per = (temp_mod - temp_def) / temp_def
        temp_per = 1 + (1 * per)
        N = len(index_lst)
        ind = np.arange(N)
        width = 0.33

        # Build plot
        fig, ax = plt.subplots(figsize=(10, 7))
        rects1 = ax.barh(ind, [1] * N, width, label='Baseline')
        rects2 = ax.barh(ind + width, temp_per, width, label='Modified')
        for i, rect in enumerate(rects2):
            height = rect.xy[1]
            ax.text(x=temp_per[i] - .05, y=height + width / 2, s=str(round(list(per)[i] * 100, 1)) + ' %', ha='right',
                    va='center')
        ax.set_yticks(ind + width / 2)
        ax.set_yticklabels(index_lst, fontsize='large')
        ax.vlines(x=1, ymin=min(ind + width / 2) - width, ymax=max(ind + width / 2) + width, colors='r',
                  linestyles=(0, (3, 3)))
        x_ticks = np.arange(start=0, stop=2, step=0.25)
        ax.set_xticks(x_ticks)
        ax.set_xticklabels([str(i) for i in x_ticks], fontsize='large')
        ax.set_xlim(left=0, right=2)
        ax.set_title(col, fontsize='xx-large')
        ax.grid(linewidth=0.5, linestyle=(0, (3, 3)), alpha=0.75)
        plt.legend(loc='upper right')
        plt.gcf().subplots_adjust(left=0.20)

        # Attachment Textbox
        labels = []
        for part in ['Muzzle', 'Barrel', 'Body', 'Underbarrel', 'Magazine', 'Handle', 'Stock']:
            for name in self._modified_stats[part].keys():
                labels.append(name)
        props = dict(boxstyle='round', alpha=0.5, facecolor='white', edgecolor='lightgrey')
        plt.text(x=1.97, y=8.7, s='\n'.join(labels), va='top', ha='right', bbox=props)

        plt.show()



# -*- coding: utf-8 -*-
"""
Created on Thu Apr 22 19:49:29 2021

@author: Peter
"""
from typing import Optional
from dataclasses import dataclass


@dataclass
class Health:
    """

    Get zombie health and armour values. Creates a list of zombie health values. Uses desired level and a health cap
    value to compute health level.

    :param level: Desired zombie level to test weapon strength at.
    :type level: int
    :param health_cap: The level when zombie health values stop increasing. In season 5 this is level 55.
        Season 4 was 85. *Optional*
    :type health_cap: int
    :param outbreak: If level is True, parameter is referencing an outbreak level, default is False. *Optional*
    :type outbreak: bool
    :param multiplier: Value used to calculate armour, default is 2. *Optional*
    :type multiplier: int
    :example:
        >>> from health_armour import Health
        >>> zombie = Health(level=20, health_cap=55, outbreak=False, multiplier=2)
        >>> zombie_health = zombie.get_health # 2,450
        >>> zombie_armour = zombie.get_armour # 1,225
    :note: The multiplier is what number you would like to divide the health number by to get armour health.
        Prior to season 4 this was half the zombies health.
    """
    def __init__(self, level: int, health_cap: Optional[int] = 55, outbreak: Optional[bool] = False,
                 multiplier: Optional[int] = 2):
        self._level = level
        self._health_cap = health_cap
        self._outbreak = outbreak
        self._multiplier = multiplier

        health_lst = []
        health_amount = 150
        for num in range(max(self._health_cap, self._level) + 1):
            if 1 < num <= 11:
                health_amount += 60
            elif 11 < num <= 16:
                health_amount += 100
            elif 16 < num <= 21:
                health_amount += 300
            elif 21 < num <= 26:
                health_amount += 750
            elif 26 < num <= 31:
                health_amount += 1400
            elif 31 < num <= 36:
                health_amount += 1750
            elif 36 < num <= health_cap:
                health_amount += 2000
            elif health_cap < num:
                pass
            health_lst.append(health_amount)

        if outbreak:
            out_health_lst = []
            for num in range(self._level):
                if num <= 1:
                    out_health_lst.append(150)
                elif 1 < num < 10:
                    out_health_lst.append(health_lst[1 + (num * 6)])
                else:
                    out_health_lst.append(health_lst[-1])

            self._hp = out_health_lst[self._level]
        else:
            self._hp = health_lst[self._level]
        self._armour = float(int(self._hp / self._multiplier))

    def __repr__(self):
        return "Zombie Health"

    @property
    def get_health(self) -> float:
        """Returns zombie health value."""
        return float(self._hp)

    @property
    def get_armour(self) -> float:
        """Returns zombie armour value. Multiplier is an int you want to divide the zombie health level by
            If 2, the armour health will be half the zombie health value."""
        return float(self._armour)

    @get_armour.setter
    def get_armour(self, val):
        """Setter for updating the multiplier and armour value."""
        self._multiplier = val
        self._armour = float(int(self._hp / self._multiplier))



from typing import List, Optional
import numpy as np
from zombie.weapon_stats import _weapon_stats_dic, _gun_names


def damage_range(weapon_dic: dict, max_range: int) -> np.ndarray:
    """Returns a np.ndarray with the damage output of a weapon at varying distances"""
    m = weapon_dic['Burst']
    dam1 = weapon_dic['Damage']
    dam2 = weapon_dic['Damage 2']
    dam3 = weapon_dic['Damage 3']
    ran1 = weapon_dic['Range']
    ran2 = weapon_dic['Range 2']
    ran = list(range(max_range))

    if weapon_dic['Weapon Type'] == 'Shotguns':
        if ran1 == 5.72:
            m2 = [1 - (i / 15) for i in range(15)]
        else:
            m1 = np.floor((ran1 / 5.72 * 20) - ran1)
            m2 = [1 - (i / m1) for i in range(m1)]

        dam_lst = [dam1] * (np.floor(ran1)) + [np.floor(dam1 * i) for i in m2]
        dam_lst = dam_lst + ([0] * len(dam_lst))
    else:
        dam_lst = []
        if ran2 == 0:
            for i in ran:
                if i < ran1:
                    dam_lst.append(dam1 * m)
                    continue
                else:
                    dam_lst.append(dam3 * m)
                    continue
        else:
            for i in ran:
                if i < ran1:
                    dam_lst.append(dam1 * m)
                    continue
                elif i < ran2:
                    dam_lst.append(dam2 * m)
                    continue
                else:
                    dam_lst.append(dam3 * m)

    return np.nan_to_num(np.array(dam_lst))


def damage_per_second(weapon_dic: dict, damage_range_arr: np.ndarray) -> np.ndarray:
    """
    Calculate the damage per second output of a specific weapon over varying distances.

    Parameters
    ----------
    weapon_dic : dict
        Dict of specific weapon stats.
    damage_range_arr : np.ndarray
        The damage at varying distances.

    Returns
    -------
    np.ndarray

    """
    return np.multiply(np.ceil(weapon_dic['Fire Rate'] / 60), damage_range_arr)


def time_to_kill(weapon_dic: dict, damage_range_arr: np.ndarray, zombie_health: float, max_range: int) -> np.ndarray:
    """
    Calculate the time to kill value of a specific weapon over varying distances.

    Parameters
    ----------
    weapon_dic : dict
        Dict of specific weapon stats.
    damage_range_arr : np.ndarray
        The damage at varying distances.
    zombie_health : float
        The zombie health value.
    max_range : int
        Max range to calculate.

    Returns
    -------
    np.ndarray

    """
    dps = damage_per_second(weapon_dic, damage_range_arr)
    temp_lst = [zombie_health / dam if dam != 0 else 0 for dam in dps]
    return np.around(np.add(temp_lst, np.divide(range(max_range), weapon_dic['Velocity'])), 2)


def shots_to_kill(damage_range_arr: np.ndarray, zombie_health: float) -> np.ndarray:
    """
    Calculate the shots to kill value of a specific weapon over varying distances.

    Parameters
    ----------
    damage_range_arr : np.ndarray
        The damage at varying distances.
    zombie_health : float
        The zombie health value.

    Returns
    -------
    np.ndarray

    """
    return np.around([zombie_health / dam if dam != 0 else 0 for dam in damage_range_arr], 0)


def shoot_time(weapon_dic: dict) -> float:
    """
    Calculate the time spent shooting in seconds.

    Parameters
    ----------
    weapon_dic : dict
        Dict of specific weapon stats.

    Returns
    -------
    float

    """
    return round(weapon_dic['Mag Capacity'] / (weapon_dic['Fire Rate'] / 60), 2)


def shoot_reload_ratio(weapon_dic: dict) -> float:
    """
    Calculate the shoot to reload ratio value of a specific weapon.

    This is used to see how long a weapon can output damage before having to be halted to reload.
    A weapon like the Bullfrog has a high shoot to reload ratio due to its large magazine size.

    Parameters
    ----------
    weapon_dic : dict
        Dict of specific weapon stats.

    Returns
    -------
    float

    """
    return 1 - round(weapon_dic['Reload'] / shoot_time(weapon_dic=weapon_dic), 2)


def damage_per_max_ammo(weapon_dic: dict, damage_range_arr: np.ndarray) -> np.ndarray:
    """
    Calculate the damage per max ammo at varying distances.

    This assumes you were to fire all ammo in the gun and in reserve at a set distance.

    Parameters
    ----------
    weapon_dic : dict
        Dict of specific weapon stats.
    damage_range_arr : np.ndarray
        The damage at varying distances.

    Returns
    -------
    np.ndarray

    """
    return np.multiply(np.divide(weapon_dic['Ammo Capacity'], weapon_dic['Burst']), damage_range_arr)


def damage_per_clip(damage_range_arr: np.ndarray, shots_average: np.ndarray) -> np.ndarray:
    """
    Calculate the damage per clip value of a specific weapon over varying distances.

    Parameters
    ----------
    damage_range_arr : np.ndarray
        The damage at varying distances.
    shots_average : np.ndarray
        An np.ndarray() with expected damage output based on accuracy and critical hit percentage.

    Returns
    -------
    np.ndarray

    """
    return np.around([damage_range_arr[i] * shots_average for i, j in enumerate(damage_range_arr)], 0)


def movement_ratio(weapon_dic: dict, weapon_names: Optional[List[str]] = None) -> float:
    """
    Calculate the movement ratio for a specific weapon.

    This is used to compare the mobility of weapon against others. The closer to 1 the better the weapons overall
    movement attributes.

    Parameters
    ----------
    weapon_dic : dict
        Dict of specific weapon stats.
    weapon_names: list
        List of weapon names, default is None. *Optional*

    Returns
    -------
    float

    """
    lst = {'Movement Speed': [0, 0, 0],
           'Sprinting Speed': [0, 0, 0],
           'Shooting Speed': [0, 0, 0],
           'Sprint to Fire': [0, 0, 0],
           'Aim Walking': [0, 0, 0],
           'ADS': [0, 0, 0]}

    if weapon_names is None:
        weapon_names = _gun_names
    else:
        weapon_names = weapon_names

    for i in lst.keys():
        temp_lst = []
        for weapon_name in weapon_names:
            if i == 'Movement Speed':
                temp_lst.append(_weapon_stats_dic[weapon_name].movement)
                continue
            elif i == 'Sprinting Speed':
                temp_lst.append(_weapon_stats_dic[weapon_name].sprint)
                continue
            elif i == 'Shooting Speed':
                temp_lst.append(_weapon_stats_dic[weapon_name].shoot_speed)
                continue
            elif i == 'Sprint to Fire':
                temp_lst.append(_weapon_stats_dic[weapon_name].sprint_to_fire)
                continue
            elif i == 'Aim Walking':
                temp_lst.append(_weapon_stats_dic[weapon_name].aim_walking)
                continue
            elif i == 'ADS':
                temp_lst.append(_weapon_stats_dic[weapon_name].ads)
                continue

        lst[i] = [min(temp_lst), max(temp_lst), weapon_dic[i]]

    ratio = round(sum([(lst[i][2] - lst[i][0]) / (lst[i][1] - lst[i][0]) for i in lst.keys()]) / len(lst.keys()), 2)

    return ratio


def control_ratio(weapon_dic: dict, weapon_names: Optional[List[str]] = None) -> float:
    """
    Calculate the control ratio for a specific weapon.

    This is used to compare the controllability of weapon against others. The closer to 1 the better the weapons
    overall control attributes.

    Parameters
    ----------
    weapon_dic : dict
        Dict of specific weapon stats.
    weapon_names: list
        List of weapon names, default is None. *Optional*

    Returns
    -------
    float

    """
    lst = {'Vertical Recoil': [0, 0, 0],
           'Horizontal Recoil': [0, 0, 0],
           'Centering Speed': [0, 0, 0],
           'Hip Fire': [0, 0, 0]}

    if weapon_names is None:
        weapon_names = _gun_names
    else:
        weapon_names = weapon_names

    for i in lst.keys():
        temp_lst = []
        for weapon_name in weapon_names:
            if i == 'Vertical Recoil':
                temp_lst.append(_weapon_stats_dic[weapon_name].vertical_recoil)
                continue
            if i == 'Horizontal Recoil':
                temp_lst.append(_weapon_stats_dic[weapon_name].horizontal_recoil)
                continue
            if i == 'Centering Speed':
                temp_lst.append(_weapon_stats_dic[weapon_name].centering_speed)
                continue
            if i == 'Hip Fire':
                temp_lst.append(_weapon_stats_dic[weapon_name].hip_fire)
                continue
        lst[i] = [min(temp_lst), max(temp_lst), weapon_dic[i]]

    ratio = 1 - round(
        sum([(lst[i][2] - lst[i][0]) / (lst[i][1] - lst[i][0]) for i in lst.keys()]) / len(lst.keys()), 2)

    return ratio


def drop_off_ratio(weapon_dic: dict) -> float:
    """
    Calculate the damage drop off ratio for a specific weapon.

    This is used to compare the how much damage is lost as targets get farther away.The closer to 0 the better
    the weapons overall control attributes. A higher value means the damage starts out strong build falls off.

    Parameters
    ----------
    weapon_dic : dict
        Dict of specific weapon stats.

    Returns
    -------
    float

    """
    return 1 + round((weapon_dic['Damage 3'] - weapon_dic['Damage']) / weapon_dic['Damage'], 2)



from dataclasses import dataclass
from typing import Optional, List
import zombie.weapon_stats as w


_stats = {'XM4': w.Xm4,
          'AK47': w.Ak47,
          'Krig': w.Krig,
          'QBZ': w.Qbz,
          'FFAR': w.Ffar,
          'Groza': w.Groza,
          'FARA': w.Fara,
          'C58': w.C58,
          'EM2': w.Em2,
          'MP5': w.Mp5,
          'Milano': w.Milano,
          'AK74u': w.Ak74u,
          'KSP': w.Ksp,
          'Bullfrog': w.Bullfrog,
          'Mac 10': w.Mac10,
          'LC10': w.Lc10,
          'PPSH': w.Ppsh,
          'OTS': w.Ots9,
          'Tec9': w.Tec9,
          'Type 63': w.Type63,
          'M16': w.M16,
          'AUG': w.Aug,
          'DMR': w.Dmr,
          'CARV': w.Carv,
          'Stoner': w.Stoner,
          'RPD': w.Rpd,
          'M60': w.M60,
          'Pellington': w.Pelington,
          'LW3': w.Lw3,
          'M82': w.M82,
          'Swiss': w.Swiss,
          '1911': w.N1911,
          'Magnum': w.Magnum,
          'Diamatti': w.Diamatti,
          'AMP': w.Amp,
          'Hauer': w.Hauer,
          'Gallo': w.Gallo,
          'Streetsweeper': w.Streetsweeper
          }
_gun_names = list(_stats.keys())

rare_level = {'common': 1.00, 'green': 1.50, 'blue': 3.00, 'purple': 6.00, 'orange': 12.00}
pack_level = {'0': 1.00, '1': 2.00, '2': 4.00, '3': 8.00}


@dataclass
class ModifiedWeapon:

    def __init__(self, weapon_name: str, nickname: Optional[str] = None, equipped_attachments: Optional[dict] = None,
                 rarity: Optional[str] = None, pap: Optional[str] = None, accuracy: Optional[float] = None,
                 critical: Optional[float] = None):
        d = _stats[weapon_name]
        self._default_stats = {'Name': d.name,
                                 'Temp Name': d.temp_name,
                                 'Damage': d.damage_1,
                                 'Damage 2': d.damage_2,
                                 'Damage 3': d.damage_3,
                                 'Range': d.range_1,
                                 'Range 2': d.range_2,
                                 'Fire Rate': d.fire_rate,
                                 'Velocity': d.velocity,
                                 'Armour Damage': d.armour_damage,
                                 'Melee Quickness': d.melee,
                                 'Movement Speed': d.movement,
                                 'Sprinting Speed': d.sprint,
                                 'Shooting Speed': d.shoot_speed,
                                 'Sprint to Fire': d.sprint_to_fire,
                                 'Aim Walking': d.aim_walking,
                                 'ADS': d.ads,
                                 'Vertical Recoil': d.vertical_recoil,
                                 'Horizontal Recoil': d.horizontal_recoil,
                                 'Centering Speed': d.centering_speed,
                                 'Idle Sway': d.idle_sway,
                                 'Flinch': d.flinch,
                                 'Hip Fire': d.hip_fire,
                                 'Mag Capacity': d.mag_capacity,
                                 'Reload': d.reload,
                                 'Starting Ammo': d.starting_ammo,
                                 'Ammo Capacity': d.ammo_capacity,
                                 'Burst': d.burst,
                                 'PAP Burst': d.pap_burst,
                                 'Long Shot': d.long_shot,
                                 'Crit': d.critical,
                                 'Long': d.long,
                                 'Close': d.close,
                                 'Salvage': d.salvage,
                                 'Equipment': d.equipment,
                                 'Attachments': d.attachments,
                                 'Weapon Type': d.weapon_type,
                                 'Accuracy': d.gun_acc_value,
                                 'Critical': d.gun_critical_value,
                                 'Pack': d.pack,
                                 'Rare': d.rare,
                                 'Muzzle': d.muzzle,
                                 'Barrel': d.barrel,
                                 'Body': d.body,
                                 'Underbarrel': d.under_barrel,
                                 'Magazine': d.magazine,
                                 'Handle': d.handle,
                                 'Stock': d.stock,
                                 }

        self._modified_stats = self._default_stats

        if nickname is not None:
            self._modified_stats['Temp Name'] = nickname

        if accuracy is not None:
            self._modified_stats['Accuracy'] = accuracy

        if critical is not None:
            self._modified_stats['Critical'] = critical

        if rarity is not None:
            self._modified_stats['Rare'] = rarity
            self._modified_stats['Damage'] = self._modified_stats['Damage'] * self.rare_level[rarity]
            self._modified_stats['Damage 2'] = self._modified_stats['Damage 2'] * self.rare_level[rarity]
            self._modified_stats['Damage 3'] = self._modified_stats['Damage 3'] * self.rare_level[rarity]

        if pap is not None:
            if pap != '0':
                self._modified_stats['Pack'] = pap
                self._modified_stats['Damage'] = self._modified_stats['Damage'] * self.pack_level[pap] * \
                                                 self._modified_stats['PAP Burst']
                self._modified_stats['Damage 2'] = self._modified_stats['Damage 2'] * self.pack_level[pap] * \
                                                   self._modified_stats['PAP Burst']
                self._modified_stats['Damage 3'] = self._modified_stats['Damage 3'] * self.pack_level[pap] * \
                                                   self._modified_stats['PAP Burst']
                self._modified_stats['Mag Capacity'] = self._modified_stats['Mag Capacity'] / \
                                                       self._modified_stats['PAP Burst']

        temp_dic = self._apply_multipliers(weapon_multi=self.get_weapon_classes, perk_multi=self.get_perk_classes,
                                           weapon_dic=self._modified_stats)

        if equipped_attachments is not None:
            effect_lst = self._get_attachments(weapon_dic=temp_dic, equipped_dic=equipped_attachments)
            final_dic = self._apply_attachments(weapon_dic=temp_dic, attachment_lst=effect_lst)

            for part in ['Muzzle', 'Barrel', 'Body', 'Underbarrel', 'Magazine', 'Handle', 'Stock']:
                if part in equipped_attachments.keys():
                    final_dic[part] = {equipped_attachments[part]: final_dic[part][equipped_attachments[part]]}
                else:
                    final_dic[part] = {'None': 'None'}
            temp_dic = final_dic

        self._modified_stats = temp_dic

    def __repr__(self):
        return self._default_stats['Name']

    @property
    def default_stats(self) -> dict:
        return self._default_stats

    @property
    def modified_stats(self) -> dict:
        return self._modified_stats



# -*- coding: utf-8 -*-
"""
Created on Thu Apr 22 21:08:22 2021

@author: Peter
"""
from typing import List, Optional
from dataclasses import dataclass
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
import random
from zombie.processor import DamageProfile, ModifiedWeapon
from zombie.health_armour import Health
from zombie.base import damage_range, damage_per_second, damage_per_clip, damage_per_max_ammo, drop_off_ratio
from zombie.base import shoot_reload_ratio, time_to_kill, shoot_time, shots_to_kill, control_ratio, movement_ratio
random.seed(1)

col_lst = ['Name', 'Temp Name', 'Damage', 'Damage 2', 'Damage 3', 'Range',
           'Range 2', 'Fire Rate', 'Velocity', 'Armour Damage', 'Melee Quickness',
           'Movement Speed', 'Sprinting Speed', 'Shooting Speed', 'Sprint to Fire',
           'Aim Walking', 'ADS', 'Vertical Recoil', 'Horizontal Recoil',
           'Centering Speed', 'Idle Sway', 'Flinch', 'Hip Fire', 'Mag Capacity',
           'Reload', 'Ammo Capacity', 'Accuracy', 'Critical', 'Pack', 'Rare', 'Shooting Time',
           'Shoot To Reload Ratio', 'Movement Ratio', 'Control Ratio', 'Drop Off Ratio']


@dataclass
class Analyze:
    """

    Constructs  dicts, pd.DataFrames and visualizations for comparing various weapons.

    :param damage_profile: DamageProfile class object.
    :type damage_profile: DamageProfile
    :param zombie_info: Health class object.
    :type zombie_info: Health
    :param weapon_dic_lst: List of weapons and weapon info.
    :type weapon_dic_lst: dict
    :param columns_lst: Select columns to highlight, default is None. *Optional*
    :type columns_lst: List[str]
    :param x_limit: X limit on the graphs, default is 75. *Optional*
    :type x_limit: int
    :param viz_index_point: Reference point on the graphs, default is 40. *Optional*
    :type viz_index_point: int
    :param save_image: If True, will save the plots, default is False. *Optional*
    :type save_image: bool
    :example:
    >>> from zombie.processor import DamageProfile
    >>> from zombie.analysis import Analyze
    >>> weapon_class_levels = {'Launcher': '5', 'Special': '5', 'Smg': '5', 'Shotgun': '5', 'Pistol': '5',
    >>>                        'Marksman': '5', 'Sniper': '5', 'Lmg': '5', 'Assault': '5', 'Melee': '5'}
    >>> perk_class_levels = {'speed': '5', 'stamin up': '5', 'deadshot': '5', 'death_perception': '5'}
    >>> damage_profile = DamageProfile(weapon_class_levels=weapon_class_levels, perk_class_levels=perk_class_levels,
    >>>                                max_range=100)
    >>> analysis = Analyze(damage_profile=damage_profile)

    """
    def __init__(self, damage_profile: DamageProfile, zombie_info: Health, weapon_dic_lst: List[dict],
                 columns_lst: Optional[List[str]] = None, x_limit: Optional[int] = 100,
                 viz_index_point: Optional[int] = 3, save_image: Optional[bool] = False):
        self._DamageProfile = damage_profile
        self._perk_dict = damage_profile.get_perk_classes
        self._weapon_type_dict = damage_profile.get_weapon_classes
        self._max_range = self._DamageProfile.max_range

        self._zombie_health = zombie_info.get_health
        self._zombie_armour = zombie_info.get_armour
        self._ModifiedWeapons = self._process_multi(weapon_dic_lst=weapon_dic_lst)
        self._compare_info_for_plots = self._compare_multi(weapon_dic_lst=self._ModifiedWeapons,
                                                           zombie_health=self._zombie_health,
                                                           zombie_armour=self._zombie_armour,
                                                           for_viz=True)

        if columns_lst is None:
            self._col_lst = col_lst
        else:
            self._col_lst = col_lst

        self._x_limit = x_limit
        self._viz_index_point = viz_index_point
        self._save_image = save_image

    def __repr__(self):
        return "Analysis"

    @property
    def weapons_lst(self) -> List[ModifiedWeapon]:
        """Returns a list of ModifiedWeapons"""
        return self._ModifiedWeapons

    @property
    def weapons_modified_df(self) -> pd.DataFrame:
        """Returns a DataFrame of weapon stats"""
        data_n = []
        for dic in self._ModifiedWeapons:
            dic = dic.modified_stats
            dic['Shooting Time'] = shoot_time(weapon_dic=dic)
            dic['Shoot To Reload Ratio'] = shoot_reload_ratio(weapon_dic=dic)
            dic['Movement Ratio'] = movement_ratio(weapon_dic=dic)
            dic['Control Ratio'] = control_ratio(weapon_dic=dic)
            dic['Drop Off Ratio'] = drop_off_ratio(weapon_dic=dic)
            data_n.append(dic)

        return pd.DataFrame(data_n)[self._col_lst].set_index('Temp Name').T

    @property
    def weapons_default_df(self) -> pd.DataFrame:
        """Returns a DataFrame of weapon stats"""
        data_n = []
        for dic in self._ModifiedWeapons:
            dic = dic.default_stats
            dic['Shooting Time'] = shoot_time(weapon_dic=dic)
            dic['Shoot To Reload Ratio'] = shoot_reload_ratio(weapon_dic=dic)
            dic['Movement Ratio'] = movement_ratio(weapon_dic=dic)
            dic['Control Ratio'] = control_ratio(weapon_dic=dic)
            dic['Drop Off Ratio'] = drop_off_ratio(weapon_dic=dic)
            data_n.append(dic)
        return pd.DataFrame(data_n)[self._col_lst].set_index('Temp Name').T

    def _process(self, weapon_dic: dict) -> ModifiedWeapon:
        """Create a ModifiedWeapon Class for a weapon with all stats. Including attachments,
        perk and weapon tier effects"""
        new_weapon_dic = {}
        for i in ['weapon', 'nickname', 'equipped_attachments', 'rarity', 'pap', 'accuracy', 'critical']:
            if i in weapon_dic.keys():
                new_weapon_dic[i] = weapon_dic[i]
            else:
                new_weapon_dic[i] = None

        return ModifiedWeapon(weapon_name=new_weapon_dic['weapon'], nickname=new_weapon_dic['nickname'],
                              equipped_attachments=new_weapon_dic['equipped_attachments'],
                              rarity=new_weapon_dic['rarity'], pap=new_weapon_dic['pap'],
                              accuracy=new_weapon_dic['accuracy'], critical=new_weapon_dic['critical'],
                              weapon_class_levels=self._DamageProfile.get_weapon_classes,
                              perk_class_levels=self._DamageProfile.get_perk_classes)

    def _process_multi(self, weapon_dic_lst: List[dict]) -> List[ModifiedWeapon]:
        """Create a List[ModifiedWeapon] for multiple weapons with all stats included in process()"""
        if len(weapon_dic_lst) > 6:
            print(), print('If more than 6 guns included legends will overlap.')
        return [self._process(weapon_dic=i) for i in weapon_dic_lst]

    def _mu_shots(self, weapon_dic: dict) -> np.ndarray:
        """Returns an np.ndarray of multipliers based on accuracy and critical hit percentages"""
        ac = weapon_dic['Accuracy']
        dis = self._DamageProfile.hits
        ind = np.cumsum([1 if i <= ac else 0 for i in dis])

        if self._DamageProfile.consecutive:
            con = [0, 1.00, 1.02, 1.04, 1.06, 1.08, 1.10, 1.12, 1.14, 1.16, 1.18, 1.20] + ([1.20] * 40)
        else:
            con = [1.00] * 50

        con_lst_master = []
        con_lst = [1]
        count = 0
        for i, j in enumerate(ind):
            if j - 1 is ind[i - 1]:
                count += 1
                con_lst.append(count)
            else:
                con_lst_master.append(con_lst)
                count, con_lst = 0, [0]

        con_lst_master_n = sum(con_lst_master, [])
        con_shots = [con[i] for i in con_lst_master_n]
        crit1 = weapon_dic['Critical']
        headshots = random.choices([0, 1], k=len(con_shots), weights=[1 - crit1, crit1])
        headshot_multi = weapon_dic['Crit']
        temp_lst = list(
            np.around([con_shots[i] * headshot_multi if j == 1.00 else 1.00 for i, j in enumerate(headshots)], 2))
        arr = np.array(temp_lst + ([0] * (2500 - len(temp_lst))))
        mc = range(int(weapon_dic['Mag Capacity']))
        average_shots = np.ceil(np.mean([sum([arr[k - q] for q in mc]) for k, l in enumerate(arr)]))
        return average_shots

    def _check_multi(self, damage_range_arr: np.ndarray, weapon_dic: dict) -> np.ndarray:
        close = weapon_dic['Close']
        long = weapon_dic['Long']
        long_shot = weapon_dic['Long Shot']
        t = list(damage_range_arr)
        temp_lst = []
        for k, j in enumerate(t):
            if k <= 4:
                temp_lst.append(j * close)
            elif k >= long_shot:
                temp_lst.append(j * long)
            else:
                temp_lst.append(j)

        return np.array(temp_lst)

    def _compare(self, weapon_dic: dict, zombie_health: Optional[float], zombie_armour: Optional[float],
                 for_viz: Optional[bool] = False):
        """
        Constructs the dict or pd.DataFrame with all comparison stats and metrics.

        Parameters
        ----------
        weapon_dic : dict
            Dict of specific weapon stats.
        zombie_health : float
            The zombie health value.
        zombie_armour : float, default None
            The zombie health value.
        for_viz : bool, default False
            Whether or not the data will be used to create plots. If False the function will return a dict.
            If True the function will return a pd.DataFrame.

        Returns
        -------
        dict or pd.DataFrame

        """
        if zombie_health is None:
            zombie_health = self._zombie_health
        if zombie_armour is None:
            zombie_armour = self._zombie_armour

        mu = self._mu_shots(weapon_dic)
        _dr = self._check_multi(damage_range_arr=damage_range(weapon_dic=weapon_dic, max_range=self._max_range),
                                weapon_dic=weapon_dic)
        if zombie_armour:
            zombie_health = zombie_health + zombie_armour
            _dr = np.multiply(np.divide(_dr, 4), weapon_dic['Armour Damage'])

        temp_dic = {'Damage Per Clip': damage_per_clip(damage_range_arr=_dr, shots_average=mu),
                    'Damage Per Second': damage_per_second(weapon_dic=weapon_dic, damage_range_arr=_dr),
                    'Time To Kill': time_to_kill(weapon_dic=weapon_dic, damage_range_arr=_dr,
                                                 zombie_health=zombie_health, max_range=self._max_range),
                    'Shots To Kill': shots_to_kill(damage_range_arr=_dr, zombie_health=zombie_health),
                    'Damage Per Max Ammo': damage_per_max_ammo(weapon_dic=weapon_dic,  damage_range_arr=_dr)
                    }

        if for_viz:
            return pd.DataFrame.from_dict(temp_dic)
        else:
            temp_dic['Temp Name'] = weapon_dic['Temp Name']
            temp_dic['Shooting Time'] = shoot_time(weapon_dic=weapon_dic)
            temp_dic['Shoot To Reload Ratio'] = shoot_reload_ratio(weapon_dic=weapon_dic)
            temp_dic['Movement Ratio'] = movement_ratio(weapon_dic=weapon_dic)
            temp_dic['Control Ratio'] = control_ratio(weapon_dic=weapon_dic)
            temp_dic['Drop Off Ratio'] = drop_off_ratio(weapon_dic=weapon_dic)
            return temp_dic

    def _compare_multi(self, weapon_dic_lst: List[ModifiedWeapon], zombie_health: Optional[float],
                       zombie_armour: Optional[float] = None, for_viz: Optional[bool] = False) -> dict:
        """
        Constructs the dict with all comparison stats and metrics; for multiple weapons.

        Parameters
        ----------
        weapon_dic_lst : List[dict]
            Dict of specific weapon stats.
        zombie_health : float
            The zombie health value.
        zombie_armour : float, default None
            The zombie health value.
        for_viz : bool, default False
            Whether or not the data will be used to create plots. If False the function will return a dict.
            If True the function will return a pd.DataFrame.

        Returns
        -------
        dict

        """
        if zombie_health is None:
            zombie_health = self._zombie_health
        if zombie_armour is None:
            zombie_armour = self._zombie_armour

        return {i.modified_stats['Name']: self._compare(weapon_dic=i.modified_stats,
                                                        zombie_health=zombie_health,
                                                        zombie_armour=zombie_armour,
                                                        for_viz=for_viz) for i in weapon_dic_lst}

    def viz(self, weapon_df_dic: dict, keyword: str, x_limit: Optional[int] = 75, ind: Optional[int] = 40,
            save_image: Optional[bool] = False) -> None:
        """
        Constructs a plot related to a phrase for comparing weapons.

        Parameters
        ----------
        weapon_df_dic : dict
            Dict of specific weapon stats. {MP5: mp5 DataFrame, ...}
        keyword : str
            Phrase used to denote a plot type.
            ['Damage Per Max Ammo', 'Damage Per Clip', 'Damage Per Second', 'Time To Kill', 'Shots To Kill']
        x_limit : int
            Maximum limit of x axis in the plot.
        ind : int
            Index used to populate the Value legend.
        save_image :  bool, default False
            If True will save the plot to the working directory.

        Returns
        -------
        None

        """
        n = 0
        df = pd.DataFrame()
        for i in weapon_df_dic.keys():
            df[i] = weapon_df_dic[i][keyword]
            if keyword in weapon_df_dic[i].columns:
                n += 1

        cmap = [plt.get_cmap('viridis')(1. * i / n) for i in range(n)]

        if "Time" not in keyword and "Shots" not in keyword:
            if np.max(np.max(df)) > 100000:
                df = df / 1000
                keyword = keyword + ' (in 1,000s)'

        for col in df.columns:
            df[col] = np.nan_to_num(np.array(df[col]), posinf=0, neginf=0)

        if 'Time' not in keyword:
            mu = df.mean(axis=1).astype(int)
        else:
            mu = df.mean(axis=1).round(1)

        labels = []
        count = 0
        for i in df.columns:

            if 'Time' not in keyword:
                val = df[i].astype(int)
            else:
                val = df[i].round(1)

            plt.plot(val, label=i, color=cmap[count])
            labels.append(str(val.iloc[ind]))
            count += 1
        plt.plot(mu, label='Mu', color='tab:orange', linestyle='--', alpha=1)
        labels = labels + [str(round(mu.iloc[ind], 1))]
        plt.title(keyword)
        legend1 = plt.legend(labels, title='Value at ' + str(ind), loc='upper right')  # bbox_to_anchor=(1.05, .5)
        plt.legend(title='Legend', loc='lower right')  # bbox_to_anchor=(1.05, .5)
        plt.xlabel('Meters')

        if 'Damage' in keyword:
            plt.ylabel('Damage')
        elif 'Time' in keyword:
            plt.ylabel('Seconds')
        elif 'Shots' in keyword:
            plt.ylabel('Shots')
        elif 'Ratio' in keyword:
            plt.ylabel('Ratio')
        else:
            plt.ylabel('Other')

        plt.grid(linewidth=1, linestyle=(0, (5, 5)), alpha=.75)
        plt.xlim(0, x_limit)
        plt.gca().add_artist(legend1)

        if save_image:
            plt.savefig(keyword)
        plt.show()

    def viz_all(self, weapon_df_dic: dict, x_limit: Optional[int] = 75, ind: Optional[int] = 40,
                save_image: Optional[bool] = False) -> None:
        """
        Constructs a series of plots related utilizing all phrases for comparing weapons.

        Parameters
        ----------
        weapon_df_dic : dict
            Dict of specific weapon stats. {MP5: mp5 DataFrame, ...}
        x_limit : int
            Maximum limit of x axis in the plot.
        ind : int
            Index used to populate the Value legend.
        save_image :  bool, default False
            If True will save the plot to the working directory.

        Returns
        -------
        None

        """
        for keyword in ['Damage Per Max Ammo', 'Damage Per Clip', 'Damage Per Second', 'Time To Kill', 'Shots To Kill']:
            self.viz(weapon_df_dic=weapon_df_dic, keyword=keyword, x_limit=x_limit, ind=ind, save_image=save_image)

    def damage_per_second_plot(self):
        """Returns Damage Per Second Plot"""
        return self.viz(weapon_df_dic=self._compare_info_for_plots,
                        keyword='Damage Per Second',
                        x_limit=self._x_limit,
                        ind=self._viz_index_point,
                        save_image=self._save_image)

    def damage_per_max_ammo_plot(self):
        """Returns Damage Per Max Ammo Plot"""
        return self.viz(weapon_df_dic=self._compare_info_for_plots,
                        keyword='Damage Per Max Ammo',
                        x_limit=self._x_limit,
                        ind=self._viz_index_point,
                        save_image=self._save_image)

    def damage_per_clip_plot(self):
        """Returns Damage Per Clip Plot"""
        return self.viz(weapon_df_dic=self._compare_info_for_plots,
                        keyword='Damage Per Clip',
                        x_limit=self._x_limit,
                        ind=self._viz_index_point,
                        save_image=self._save_image)

    def time_to_kill_plot(self):
        """Returns Time To Kill Plot"""
        return self.viz(weapon_df_dic=self._compare_info_for_plots,
                        keyword='Time To Kill',
                        x_limit=self._x_limit,
                        ind=self._viz_index_point,
                        save_image=self._save_image)

    def shots_to_kill_plot(self):
        """Returns Shots To Kill Plot"""
        return self.viz(weapon_df_dic=self._compare_info_for_plots,
                        keyword='Shots To Kill',
                        x_limit=self._x_limit,
                        ind=self._viz_index_point,
                        save_image=self._save_image)

    def comparison(self):
        """Returns a bar plot comparing weapons"""
        dic = {}
        for i in self._ModifiedWeapons:
            dic[i.modified_stats['Temp Name']] = i._modified_values
        temp = pd.DataFrame.from_dict(dic)
        t = temp.mean(axis=1)
        tt = 1 + (temp.sub(t, axis=0).div(t, axis=0) * 1)
        n = len(temp.columns)
        cmap = [plt.get_cmap('viridis')(1. * i / n) for i in range(n)]
        colors = dict(zip(temp.columns, cmap))
        tt.plot.barh(color=colors, figsize=(10, 7), xlim=(np.min(np.min(tt)), 2), legend='reverse')
        plt.title('Comparison', fontsize='xx-large')
        plt.vlines(x=1, ymin=-1, ymax=len(temp), color='r', linestyles=(0, (3, 3)))
        plt.grid(linewidth=0.5, linestyle=(0, (3, 3)), alpha=0.75)
        plt.gcf().subplots_adjust(left=0.15)
        plt.show()
